﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Testing;

namespace OpcLabs.UnitTestLib.Testers
{
    public class EquivalenceTester : Tester
    {
        public EquivalenceTester() 
        {
            TestCategories.Add("Equality");
        }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            int i = 0;
            foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator in InstanceCreators)
            {
                string instanceName = instanceCreator.Key;
                Debug.Assert(instanceName != null);

                AddTestMethod(CreateReflexivityTestMethod(instanceName));
                if (i < 1)
                    AddTestMethod(CreateWithNullTestMethod(instanceName));

                i++;
            }

            if (GetOption("UnequalTest"))
                AddTestMethod(CreateUnequalTestMethod());
        }

        [NotNull]
        private TestMethod CreateReflexivityTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Reflexivity_{0}", instanceName), 
                () =>
                    {
                        object instance = CreateInstance(instanceName);
                        AssertUtilities.WriteLine(instance.GetHashCode());
                        // ReSharper disable EqualExpressionComparison
                        Assert.IsTrue(instance.Equals(instance));
                        // ReSharper restore EqualExpressionComparison
                    });
        }

        [NotNull]
        private TestMethod CreateUnequalTestMethod()
        {
            return new DelegatedTestMethod(
                "Unequal",
                () =>
                {
                    if (InstanceCreators.Count <= 1)
                        AssertUtilities.Inconclusive("Not enough instances available ({0}). At least two instances are needed.", InstanceCreators.Count);

                    foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator1 in InstanceCreators)
                    {
                        string instanceName1 = instanceCreator1.Key;
                        Debug.Assert(instanceName1 != null);

                        foreach (
                            KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator2 in InstanceCreators)
                        {
                            string instanceName2 = instanceCreator2.Key;
                            Debug.Assert(instanceName2 != null);

                            if (instanceName2 == instanceName1)
                                continue;

                            AssertUtilities.WriteLine($"Instance names: {instanceName1}, {instanceName2}");

                            object instance1 = CreateInstance(instanceName1);
                            AssertUtilities.WriteLine($"Hash 1: {instance1.GetHashCode()}");

                            object instance2 = CreateInstance(instanceName2);
                            AssertUtilities.WriteLine($"Hash 2: {instance2.GetHashCode()}");

                            Assert.IsFalse(instance1.Equals(instance2), "Equals method returned true but it should have returned false.");
                        }
                    }
                });
        }

        [NotNull]
        private TestMethod CreateWithNullTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "WithNull_{0}", instanceName),
                () =>
                    {
                        object instance = CreateInstance(instanceName);
                        Assert.IsFalse(instance.Equals(null));
                    });
        }
    }
}
