﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public partial class IEasyDAClientTest_Intrinsic
    {
        #region Static Private Properties

        [NotNull]
        static private DAItemDescriptor ValidItemDescriptor => new DAItemDescriptor("testItemId");

        [NotNull]
        static private DANodeDescriptor ValidNodeDescriptor => new DANodeDescriptor("testItemId");

        [NotNull]
        static private DAPropertyDescriptor ValidPropertyDescriptor => DAPropertyIds.Value;

        [NotNull]
        static private DAReadParameters ValidReadParameters => DAReadParameters.Default;

        [NotNull]
        static private DAGroupParameters ValidGroupParameters => DAGroupParameters.Default;

        [NotNull]
        static private ServerDescriptor ValidServerDescriptor => new ServerDescriptor("opcda:testServer");

        [NotNull]
        static private object ValidValue => "testValue";

        [NotNull]
        static private DAVtq ValidVtq => new DAVtq(ValidValue);

        #endregion



        #region Static Private Methods

        static private void AssertOpcException([NotNull] Action action)
        {
            try
            {
                action();
            }
            catch (OpcException opcException)
            {
                AssertUtilities.WriteLine(opcException);
                int errorCode = opcException.ErrorCode;
                AssertUtilities.WriteLine(errorCode);
                Assert.IsTrue(errorCode < 0);
                return;
            }
            Assert.Fail();
        }

        public void SubscribeMultipleItemsTest([NotNull] EasyDAItemSubscriptionArguments[] argumentsArray, bool useCallback)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine($"Use callback: {useCallback}");

            IEasyDAClient client = CreateClient();

            var itemChangedEventArgsList = new List<EasyDAItemChangedEventArgs>();
            client.ItemChanged += (sender, args) =>
            {
                AssertUtilities.WriteLine($"ItemChanged: {args}");
                itemChangedEventArgsList.Add(args);
            };

            var callbackEventArgsList = new List<EasyDAItemChangedEventArgs>();
            if (useCallback)
                foreach (EasyDAItemSubscriptionArguments arguments in argumentsArray)
                    // ReSharper disable once PossibleNullReferenceException
                    arguments.Callback = (sender, args) =>
                    {
                        AssertUtilities.WriteLine($"callback: {args}");
                        callbackEventArgsList.Add(args);
                    };

            int[] handleArray;
            try
            {
                TestSection.Act();

                handleArray = client.SubscribeMultipleItems(argumentsArray);
                Assert.IsNotNull(handleArray);
                Assert.AreEqual(3, handleArray.Length);
                foreach (int handle in handleArray)
                    Assert.IsTrue(client.IsKnownItemSubscriptionHandle(handle));
                int distinctHandleCount = handleArray.Distinct().Count();
                Assert.AreEqual(3, distinctHandleCount);

                AssertUtilities.WaitingPhase(5 * 1000);
            }
            finally
            {
                client.UnsubscribeAllItems();
            }
            AssertUtilities.WaitingPhase(3*1000);

            TestSection.Assert();

            AssertUtilities.WriteLine($"Number of events from ItemChanged: {itemChangedEventArgsList.Count}");
            if (useCallback)
                AssertUtilities.WriteLine($"Number of events from callback: {callbackEventArgsList.Count}");

            bool ErrorEventPredicate(EasyDAItemChangedEventArgs args) => (
                // ReSharper disable once MergeSequentialChecks
                (args != null) && (args.Arguments.State is int) && ((int) args.Arguments.State == 1));

            void ErrorEventAsserts(EasyDAItemChangedEventArgs args, ServerDescriptor serverDescriptor, DAItemDescriptor itemDescriptor, DAGroupParameters groupParameters, int handle)
            {
                Assert.IsNotNull(args, "Event not found");
                AssertUtilities.WriteLine(args);

                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);
                Assert.AreEqual(itemDescriptor, args.Arguments.ItemDescriptor);
                Assert.AreEqual(groupParameters, args.Arguments.GroupParameters);

                Assert.IsNotNull(args.Exception, "Exception is null");
                int errorCode = args.ErrorCode;
                AssertUtilities.WriteLine(errorCode);
                Assert.IsTrue(errorCode < 0, "Error code is non-negative");
                Assert.IsFalse(String.IsNullOrEmpty(args.ErrorMessage), "Error message is null or empty");
            }

            EasyDAItemSubscriptionArguments arguments1 = argumentsArray[1];
            Debug.Assert(arguments1 != null);

            AssertUtilities.WriteLine("Checking events from ItemChanged");
            EasyDAItemChangedEventArgs eventArgsFromItemChanged = itemChangedEventArgsList
                .FirstOrDefault(ErrorEventPredicate);
            ErrorEventAsserts(
                eventArgsFromItemChanged,
                arguments1.ServerDescriptor,
                arguments1.ItemDescriptor,
                arguments1.GroupParameters,
                handleArray[1]);

            if (useCallback)
            {
                AssertUtilities.WriteLine("Checking events from callback");
                EasyDAItemChangedEventArgs eventArgsFromCallback = callbackEventArgsList
                    .FirstOrDefault(ErrorEventPredicate);
                ErrorEventAsserts(
                    eventArgsFromCallback,
                    arguments1.ServerDescriptor,
                    arguments1.ItemDescriptor,
                    arguments1.GroupParameters,
                    handleArray[1]);
            }
        }

        #endregion



        #region Test Methods

        [ExpectedException(typeof (OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseAccessPaths(ServerDescriptor.Null, ValidNodeDescriptor);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_NullServerDescriptor_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseAccessPaths(ServerDescriptor.Null, ValidNodeDescriptor));
        }

        [ExpectedException(typeof (OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_UnknownTechnology()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseAccessPaths(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_UnknownTechnology_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseAccessPaths(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor));
        }

        [ExpectedException(typeof (OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseNodes(ServerDescriptor.Null, ValidNodeDescriptor, DABrowseParameters.Empty);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_NullServerDescriptor_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseNodes(ServerDescriptor.Null, ValidNodeDescriptor, DABrowseParameters.Empty));
        }

        [ExpectedException(typeof(OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_UnknownTechnology()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseNodes(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor, DABrowseParameters.Empty);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseNodesTest_UnknownTechnology_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseNodes(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor, DABrowseParameters.Empty));
        }

        [ExpectedException(typeof(OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseProperties(ServerDescriptor.Null, ValidNodeDescriptor);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_NullServerDescriptor_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseProperties(ServerDescriptor.Null, ValidNodeDescriptor));
        }

        [ExpectedException(typeof(OpcException))]
        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_UnknownTechnology()
        {
            IEasyDAClient client = CreateClient();
            client.BrowseProperties(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_UnknownTechnology_Details()
        {
            IEasyDAClient client = CreateClient();
            AssertOpcException(() => client.BrowseProperties(new ServerDescriptor("x-testSchema:unknown!"), ValidNodeDescriptor));
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetMultiplePropertyValuesTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            ValueResult[] valueResultArray = client.GetMultiplePropertyValues(new[]
            {
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 0), 
                new DAPropertyArguments(ServerDescriptor.Null, ValidNodeDescriptor, DAPropertyIds.Value, state: 1), 
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 2)
            });
            Assert.IsNotNull(valueResultArray);
            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            Assert.AreEqual(1, valueResult1.State);
            Exception exception1 = valueResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(valueResult1.Value);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetMultiplePropertyValuesTest_NullNodeDescriptor()
        {
            IEasyDAClient client = CreateClient();
            ValueResult[] valueResultArray = client.GetMultiplePropertyValues(new[]
            {
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 0), 
                new DAPropertyArguments(ValidServerDescriptor, DANodeDescriptor.Null, ValidPropertyDescriptor, state: 1), 
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 2)
            });
            Assert.IsNotNull(valueResultArray);
            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            Assert.AreEqual(1, valueResult1.State);
            Exception exception1 = valueResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(valueResult1.Value);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetMultiplePropertyValuesTest_NullPropertyDescriptor()
        {
            IEasyDAClient client = CreateClient();
            ValueResult[] valueResultArray = client.GetMultiplePropertyValues(new[]
            {
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 0), 
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, DAPropertyDescriptor.Null, state: 1), 
                new DAPropertyArguments(ValidServerDescriptor, ValidNodeDescriptor, ValidPropertyDescriptor, state: 2)
            });
            Assert.IsNotNull(valueResultArray);
            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            Assert.AreEqual(1, valueResult1.State);
            Exception exception1 = valueResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(valueResult1.Value);

            OperationResult operationResult0 = valueResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = valueResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            DAVtqResult[] vtqResultArray = client.ReadMultipleItems(new[]
            {
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 0), 
                new DAReadItemArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidReadParameters, state: 1), 
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 2)
            });
            Assert.IsNotNull(vtqResultArray);
            DAVtqResult vtqResult1 = vtqResultArray[1];
            Assert.IsNotNull(vtqResult1);
            Assert.AreEqual(1, vtqResult1.State);
            Exception exception1 = vtqResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(vtqResult1.Vtq);

            OperationResult operationResult0 = vtqResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = vtqResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_NullItemDescriptor()
        {
            IEasyDAClient client = CreateClient();
            DAVtqResult[] vtqResultArray = client.ReadMultipleItems(new[]
            {
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 0), 
                new DAReadItemArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidReadParameters, state: 1), 
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 2)
            });
            Assert.IsNotNull(vtqResultArray);
            DAVtqResult vtqResult1 = vtqResultArray[1];
            Assert.IsNotNull(vtqResult1);
            Assert.AreEqual(1, vtqResult1.State);
            Exception exception1 = vtqResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(vtqResult1.Vtq);

            OperationResult operationResult0 = vtqResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = vtqResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemValuesTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            ValueResult[] valueResultArray = client.ReadMultipleItemValues(new[]
            {
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 0), 
                new DAReadItemArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidReadParameters, state: 1), 
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 2)
            });
            Assert.IsNotNull(valueResultArray);
            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            Assert.AreEqual(1, valueResult1.State);
            Exception exception1 = valueResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(valueResult1.Value);

            OperationResult operationResult0 = valueResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = valueResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemValuesTest_NullItemDescriptor()
        {
            IEasyDAClient client = CreateClient();
            ValueResult[] valueResultArray = client.ReadMultipleItemValues(new[]
            {
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 0), 
                new DAReadItemArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidReadParameters, state: 1), 
                new DAReadItemArguments(ValidServerDescriptor, ValidItemDescriptor, ValidReadParameters, state: 2)
            });
            Assert.IsNotNull(valueResultArray);
            ValueResult valueResult1 = valueResultArray[1];
            Assert.IsNotNull(valueResult1);
            Assert.AreEqual(1, valueResult1.State);
            Exception exception1 = valueResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));
            Assert.IsNull(valueResult1.Value);

            OperationResult operationResult0 = valueResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = valueResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeMultipleItemsTest_NullServerDescriptor()
        {
            var argumentsArray = new[]
                {
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 0), 
                    new EasyDAItemSubscriptionArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidGroupParameters, state: 1), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 2)
                };
            SubscribeMultipleItemsTest(argumentsArray, useCallback:false);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeMultipleItemsTest_NullServerDescriptor_Callback()
        {
            var argumentsArray = new[]
                {
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 0), 
                    new EasyDAItemSubscriptionArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidGroupParameters, state: 1), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 2)
                };
            SubscribeMultipleItemsTest(argumentsArray, useCallback: true);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeMultipleItemsTest_NullItemDescriptor()
        {
            var argumentsArray = new[]
                {
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 0), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidGroupParameters, state: 1), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 2)
                };
            SubscribeMultipleItemsTest(argumentsArray, useCallback: false);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeMultipleItemsTest_NullItemDescriptor_Callback()
        {
            var argumentsArray = new[]
                {
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 0), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidGroupParameters, state: 1), 
                    new EasyDAItemSubscriptionArguments(ValidServerDescriptor, ValidItemDescriptor, ValidGroupParameters, state: 2)
                };
            SubscribeMultipleItemsTest(argumentsArray, useCallback: true);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemsTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            OperationResult[] operationResultArray = client.WriteMultipleItems(new[]
            {
                new DAItemVtqArguments(ValidServerDescriptor, ValidItemDescriptor, ValidVtq, state: 0), 
                new DAItemVtqArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidVtq, state: 1), 
                new DAItemVtqArguments(ValidServerDescriptor, ValidItemDescriptor, ValidVtq, state: 2)
            });
            Assert.IsNotNull(operationResultArray);
            OperationResult operationResult = operationResultArray[1];
            Assert.IsNotNull(operationResult);
            Assert.AreEqual(1, operationResult.State);
            Exception exception1 = operationResult.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));

            OperationResult operationResult0 = operationResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = operationResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemsTest_NullItemDescriptor()
        {
            IEasyDAClient client = CreateClient();
            OperationResult[] operationResultArray = client.WriteMultipleItems(new[]
            {
                new DAItemVtqArguments(ValidServerDescriptor, ValidItemDescriptor, ValidVtq, state: 0), 
                new DAItemVtqArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidVtq, state: 1), 
                new DAItemVtqArguments(ValidServerDescriptor, ValidItemDescriptor, ValidVtq, state: 2)
            });
            Assert.IsNotNull(operationResultArray);
            OperationResult operationResult1 = operationResultArray[1];
            Assert.IsNotNull(operationResult1);
            Assert.AreEqual(1, operationResult1.State);
            Exception exception1 = operationResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));

            OperationResult operationResult0 = operationResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = operationResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemValuesTest_NullServerDescriptor()
        {
            IEasyDAClient client = CreateClient();
            OperationResult[] operationResultArray = client.WriteMultipleItemValues(new[]
            {
                new DAItemValueArguments(ValidServerDescriptor, ValidItemDescriptor, ValidValue, state: 0), 
                new DAItemValueArguments(ServerDescriptor.Null, ValidItemDescriptor, ValidValue, state: 1), 
                new DAItemValueArguments(ValidServerDescriptor, ValidItemDescriptor, ValidValue, state: 2)
            });
            Assert.IsNotNull(operationResultArray);
            OperationResult operationResult1 = operationResultArray[1];
            Assert.IsNotNull(operationResult1);
            Assert.AreEqual(1, operationResult1.State);
            Exception exception1 = operationResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));

            OperationResult operationResult0 = operationResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = operationResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        [TestCategory("ErrorHandling")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemValuesTest_NullItemDescriptor()
        {
            IEasyDAClient client = CreateClient();
            OperationResult[] operationResultArray = client.WriteMultipleItemValues(new[]
            {
                new DAItemValueArguments(ValidServerDescriptor, ValidItemDescriptor, ValidValue, state: 0), 
                new DAItemValueArguments(ValidServerDescriptor, DAItemDescriptor.Null, ValidValue, state: 1), 
                new DAItemValueArguments(ValidServerDescriptor, ValidItemDescriptor, ValidValue, state: 2)
            });
            Assert.IsNotNull(operationResultArray);
            OperationResult operationResult1 = operationResultArray[1];
            Assert.IsNotNull(operationResult1);
            Assert.AreEqual(1, operationResult1.State);
            Exception exception1 = operationResult1.Exception;
            Assert.IsNotNull(exception1);
            AssertUtilities.WriteLine(exception1);
            Assert.IsNotInstanceOfType(exception1, typeof(OpcException));

            OperationResult operationResult0 = operationResultArray[0];
            Assert.IsNotNull(operationResult0);
            Assert.AreEqual(0, operationResult0.State);

            OperationResult operationResult2 = operationResultArray[2];
            Assert.IsNotNull(operationResult2);
            Assert.AreEqual(2, operationResult2.State);
        }

        #endregion
    }
}
