﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.EasyOpc.UA.Toolkit;
using OpcLabs.EasyOpc.UA.Toolkit.ClientServer;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance.AttributeServices
{
    public abstract partial class ReadBaseTest : EndpointTestBase
    {
        #region Private Constants

        private const int NodeCount20Plus = 50; // the number used when "20+" is mentioned in the spec

        #endregion



        #region Static Private Methods

        [NotNull]
        static private Type[] GetTestTypeArray()
        {
            var typeArray = new[]
                {
                    typeof (Boolean[]),
                    typeof (Byte[]),
                    typeof (SByte[]),
                    typeof (Byte[][]), 
                    //typeof(Char[]),   // There is no such type in UA
                    typeof (DateTime[]),
                    typeof (Double[]),
                    typeof (Single[]),
                    typeof (Guid[]),
                    typeof (Int16[]),
                    typeof (UInt16[]),
                    typeof (Int32[]),
                    typeof (UInt32[]),
                    typeof (Int64[]),
                    typeof (UInt64[]),
                    typeof (String[])
                };
            return typeArray;
        }

        static private bool IsNoServerTimestampReturnedException([NotNull] Exception exception)
        {
            var clientEngineException = exception as UAClientEngineException;
            // ReSharper disable once UseNullPropagation
            if (clientEngineException == null)
                return false;

            return clientEngineException.ErrorCode == UAClientErrorCodes.NoServerTimestampReturned;
        }

        static private bool IsNoSourceTimestampReturnedException([NotNull] Exception exception)
        {
            var clientEngineException = exception as UAClientEngineException;
            // ReSharper disable once UseNullPropagation
            if (clientEngineException == null)
                return false;

            return clientEngineException.ErrorCode == UAClientErrorCodes.NoSourceTimestampReturned;
        }

        static private bool IsServerTimestampAheadException([NotNull] Exception exception)
        {
            var serviceException = exception as UAServiceException;
            // ReSharper disable once UseNullPropagation
            if (serviceException == null)
                return false;

            if (serviceException.ServiceResult != UACodeBits.BadInvalidTimestamp)
                return false;

            if (!serviceException.Message.Contains("Server timestamp"))
                return false;

            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (!serviceException.Message.Contains("too much ahead"))
                return false;

            return true;
        }

        static private bool IsSourceTimestampAheadException([NotNull] Exception exception)
        {
            var serviceException = exception as UAServiceException;
            // ReSharper disable once UseNullPropagation
            if (serviceException == null)
                return false;

            if (serviceException.ServiceResult != UACodeBits.BadInvalidTimestamp)
                return false;

            if (!serviceException.Message.Contains("Source timestamp"))
                return false;

            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (!serviceException.Message.Contains("too much ahead"))
                return false;

            return true;
        }

        static private bool IsStaleTimestampException([NotNull] Exception exception)
        {
            var serviceException = exception as UAServiceException;
            // ReSharper disable once UseNullPropagation
            if (serviceException == null)
                return false;

            if (serviceException.ServiceResult != UACodeBits.BadInvalidTimestamp)
                return false;

            return 
                serviceException.Message.ToLower().Contains("stale") &&
                serviceException.Message.ToLower().Contains("timestamp");
        }

        #endregion






#region Private Methods

#if (NETFRAMEWORK)
        [NotNull]
        private ReadRecord AssertAndGetLastReadRecord([NotNull] UAOperationRecorder recorder)
        {
            int readRecordCount = recorder.ReadRecordList.Count;
            Assert.IsTrue(readRecordCount >= 1);
            ReadRecord readRecord = recorder.ReadRecordList[readRecordCount - 1];
            Debug.Assert(readRecord != null);
            return readRecord;
        }
#endif

        private void AssertAttributeDataType([CanBeNull] UAAttributeData attributeData, [NotNull] Type expectedValueType)
        {
            Assert.IsNotNull(attributeData);
            AssertUtilities.WriteLine($"attributeData: {attributeData}");
            bool isGoodStatus = attributeData.HasGoodStatus;
            Assert.IsTrue(isGoodStatus);
            object value = attributeData.Value;
            Assert.IsNotNull(value);
            Assert.AreEqual(expectedValueType, value.GetType());
        }

        private void AssertAttributeDataValue([CanBeNull] UAAttributeData attributeData, [CanBeNull] object expectedValue)
        {
            Assert.IsNotNull(attributeData);
            AssertUtilities.WriteLine($"attributeData: {attributeData}");
            bool isGoodStatus = attributeData.HasGoodStatus;
            Assert.IsTrue(isGoodStatus);

            object value = attributeData.Value;
            if (expectedValue == null)
            {
                Assert.IsNull(value);
                return;
            }
            Assert.IsNotNull(value);
            Type expectedType = expectedValue.GetType();
            Type valueType = value.GetType();
            if (expectedType.IsArray)
            {
                Assert.IsTrue(valueType.IsArray);
                Assert.AreEqual(expectedType, valueType);

                var expectedArray = (Array)expectedValue;
                var actualArray = (Array)value;

                Assert.IsTrue(
                    ((IStructuralEquatable)expectedArray).Equals(actualArray, EqualityComparer<Object>.Default));
                return;
            }

            if (expectedType == typeof(Single))
                Assert.IsTrue(Math.Abs((Single)value - (Single)expectedValue) < 1.175494E-38);
            else if (expectedType == typeof(Double))
                Assert.IsTrue(Math.Abs((Double)value - (Double)expectedValue) < 2.2250738585072E-308);
            else
                Assert.AreEqual(expectedValue, value);

        }

        private void AssertAttributeDataResultArray([CanBeNull] UAAttributeDataResult[] attributeDataResultArray,
            int expectedLength, bool mustBeGoodStatus)
        {
            Assert.IsNotNull(attributeDataResultArray);
            int length = attributeDataResultArray.Length;
            Assert.AreEqual(expectedLength, length);
            CollectionAssert.AllItemsAreNotNull(attributeDataResultArray);
            // ReSharper disable PossibleNullReferenceException
            Assert.IsTrue(
                attributeDataResultArray.All(result => result.Exception == null));
            Assert.IsTrue(
                attributeDataResultArray.All(result => (result.AttributeData != null) &&
                    (!mustBeGoodStatus || result.AttributeData.HasGoodStatus)));
            // ReSharper restore PossibleNullReferenceException
        }

#if (NETFRAMEWORK)
        private void AssertMaxAgeOutcome([NotNull] UAOperationRecorder recorder,
            [CanBeNull] UAAttributeData attributeData, [NotNull] Type expectedValueType,
            double expectedMaxAge)
        {
            AssertAttributeDataType(attributeData, expectedValueType);
            AssertReadRecordMaxAges(recorder, 1, expectedMaxAge);
        }
#endif

#if (NETFRAMEWORK)
        private void AssertTimestampsOutcome([NotNull] UAOperationRecorder recorder,
            [NotNull] UAAttributeData[] attributeDataArray, [NotNull] Type expectedValueType)
        {
            foreach (UAAttributeData attributeData in attributeDataArray)
                AssertAttributeDataType(attributeData, expectedValueType);

            ReadRecord readRecord = AssertAndGetLastReadRecord(recorder);
            Assert.AreEqual(TimestampsToReturn.Both, readRecord.TimestampsToReturn);

            int expectedCount = attributeDataArray.Length;
            ReadValueIdCollection readValueIdCollection = readRecord.NodesToRead;
            Assert.IsNotNull(readValueIdCollection);
            Assert.AreEqual(expectedCount, readValueIdCollection.Count);
            DataValueCollection dataValueCollection = readRecord.Results;
            Assert.IsNotNull(dataValueCollection);
            Assert.AreEqual(expectedCount, dataValueCollection.Count);
            for (int i = 0; i < expectedCount; i++)
            {
                ReadValueId readValueId = readValueIdCollection[i];
                Assert.IsNotNull(readValueId);
                Assert.IsNotNull(readValueId.NodeId); // IMPROVE: check if it is valid and correct
                Assert.IsNull(readValueId.IndexRange);
                Assert.AreEqual(Attributes.Value, readValueId.AttributeId);

                DataValue dataValue = dataValueCollection[i];
                Assert.IsNotNull(dataValue);
                // IMPROVE: actually there is no guarantee that indices match, but it works currently
                UAAttributeData attributeData = attributeDataArray[i];
                Debug.Assert(attributeData != null);
                Assert.AreEqual(dataValue.ServerTimestamp, attributeData.ServerTimestamp);
                Assert.AreEqual(dataValue.SourceTimestamp, attributeData.SourceTimestamp);
            }
        }
#endif

#if (NETFRAMEWORK)
        private void AssertReadRecordMaxAges([NotNull] UAOperationRecorder recorder, int expectedCount, double expectedMaxAge)
        {
            ReadRecord readRecord = AssertAndGetLastReadRecord(recorder);
            ReadValueIdCollection readValueIdCollection = readRecord.NodesToRead;
            Assert.IsNotNull(readValueIdCollection);
            Assert.AreEqual(expectedCount, readValueIdCollection.Count);
            for (int i = 0; i < expectedCount; i++)
            {
                ReadValueId readValueId = readValueIdCollection[i];
                Assert.IsNotNull(readValueId);
                Assert.IsNotNull(readValueId.NodeId); // IMPROVE: check if it is valid and correct
                Assert.IsNull(readValueId.IndexRange);
                Assert.AreEqual(Attributes.Value, readValueId.AttributeId);
                Assert.AreEqual(expectedMaxAge, readRecord.MaxAge);
            }
        }
#endif

#endregion
    }
}
