﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.IO.Utilities;
using OpcLabs.BaseLib.Portable;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    static class Ctt
    {
        #region Static Public Properties

        [Category("Classification")]
        static public bool IsScriptRunning => ScriptRunner != null;

        static public uint ScriptCrc { get; private set; }

        #endregion


        
        #region Static Public Methods

        static public void ClaimScriptBuggyVersion(UInt32 crc)
        {
            if (IsScriptRunning)
                // ReSharper disable once LocalizableElement
                throw new CustomException("You can only claim buggy script version when the CTT script is not running.");
            BuggyScriptVersions.Add(crc);
        }

        static public ScriptRunner RunScript([NotNull] string projectName, [NotNull] string scriptName)
        {
            Debug.Assert(projectName != null);
            Debug.Assert(scriptName != null);

            return RunScript(projectName, scriptName, modifyRunnerAction: null);
        }

        static public ScriptRunner RunScript([NotNull] string projectName, [NotNull] string scriptName,
            [CanBeNull] Action<ScriptRunner> modifyRunnerAction)
        {
            Debug.Assert(projectName != null);
            Debug.Assert(scriptName != null);

            if (IsScriptRunning)
                // ReSharper disable once LocalizableElement
                throw new CustomException("The CTT script is already running.");

            AssertUtilities.WriteSeparatorLine();

            //Process[] pname = Process.GetProcessesByName("uacompliancetest");
            //if (pname.Length == 0)
            //    Assert.Inconclusive("Compliance Test Tool is not running");

            string assemblyLocation = Assembly.GetExecutingAssembly().Location;
            string assemblyDirectory = Path.GetDirectoryName(assemblyLocation);
            Trace.Assert(assemblyDirectory != null);
            string projectsBase = assemblyDirectory;

            ProjectFile = Path.Combine(projectsBase, projectName);
            string projectDirectory = Path.GetDirectoryName(ProjectFile);
            Trace.Assert(projectDirectory != null);
            ScriptFile = Path.Combine(Path.Combine(projectDirectory, "maintree"), scriptName);
            ResultFile = Path.Combine(assemblyDirectory, "uacompliancetest.result.xml");

            if (!File.Exists(ProjectFile))
                Assert.Inconclusive("CTT project file \"{0}\" does not exist.", ProjectFile);
            if (!File.Exists(ScriptFile))
                Assert.Inconclusive("CTT script file \"{0}\" does not exist.", ScriptFile);

            uint crc = FileUtilities.ComputeCrc32(ScriptFile);
            AssertUtilities.WriteLine($"CTT script CRC: 0x{crc:x8}");
            ScriptCrc = crc;
            if (BuggyScriptVersions.Contains(crc))
                Assert.Inconclusive("This CTT script version is known to contain a bug that causes it to fail the test case.");

            ScriptRunner = new ScriptRunner(ProjectFile, ScriptFile, ResultFile);
            ScriptRunner.Closed += ScriptRunnerOnClosed;
            modifyRunnerAction?.Invoke(ScriptRunner);

            ScriptRunner.Run();
            return ScriptRunner;
        }

        #endregion



        #region Static Private Properties

        [NotNull]
        static private readonly HashSet<UInt32> BuggyScriptVersions = new HashSet<uint>();

        [CanBeNull]
        static private string ProjectFile { get; set; }

        [CanBeNull]
        static private string ResultFile { get; set; }

        [CanBeNull]
        static private string ScriptFile { get; set; }

        [CanBeNull]
        static private ScriptRunner ScriptRunner { get; set; }

        #endregion



        #region Static Private Methods

        static private void ScriptRunnerOnClosed(object sender, EventArgs eventArgs)
        {
            Debug.Assert(eventArgs != null);
            Debug.Assert(ScriptFile != null);

            AssertUtilities.WriteLine(
                $"The CTT script runner closed, the script file is available from file://{ScriptFile.Replace('\\', '/').Replace(" ", "%20")}");
            ScriptRunner = null;
            BuggyScriptVersions.Clear();
        }

        #endregion
    }
}
