﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.InformationModel;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    public abstract class EndpointTestBase
    {
        #region Protected Constants

        protected const int MillisecondsToCollectMultipleNotifications = 5 * 1000;
        protected const int MillisecondsToCollectSingleNotification = 3 * 1000;

        #endregion



        #region Protected Constructors

        protected EndpointTestBase()
        {
#if (NETFRAMEWORK)
            CttProjectName = "";
            EndpointDescriptorCtt = "opc.tcp://localhost:4842";    // through CTT
#endif
            EndpointDescriptorDirect = new UAEndpointDescriptor();
        }

#endregion



#region Static Protected Methods

        static protected bool IsClientEngineException([NotNull] Exception exception, int errorCode)
        {
            var clientEngineException = exception as UAClientEngineException;
            // ReSharper disable once UseNullPropagation
            if (clientEngineException == null)
                return false;

            return clientEngineException.ErrorCode == errorCode;
        }

#endregion



#region Public Properties

#if (NETFRAMEWORK)
        [NotNull]
        public string CttProjectName { get; protected set; }
#endif

        [NotNull]
        public UAEndpointDescriptor EndpointDescriptor =>
#if (NETFRAMEWORK)
            Ctt.IsScriptRunning ? EndpointDescriptorCtt : 
#endif
                EndpointDescriptorDirect;

        [NotNull]
        public UAEndpointDescriptor EndpointDescriptorDirect { get; protected set; }

#if (NETFRAMEWORK)
        [NotNull]
        public UAEndpointDescriptor EndpointDescriptorCtt { get; protected set; }
#endif

#endregion



#region Public Methods

        [TestInitialize]
        public void TestSetup()
        {
            AssertUtilities.Phase("Closing all sessions");
            EasyUAClient.CloseAll();
        }

#endregion



#region Protected Methods

        // ReSharper disable UnusedParameter.Local
        protected void AssertNoFailures([NotNull] IEnumerable<EasyUAMonitoredItemChangedEventArgs> eventArgsEnumerable)
        // ReSharper restore UnusedParameter.Local
        {
            // ReSharper disable PossibleNullReferenceException
            Assert.IsFalse(eventArgsEnumerable.Any(eventArgs => eventArgs.Exception != null));
            // ReSharper restore PossibleNullReferenceException
        }

        protected void AssertServiceException(
            [CanBeNull] Exception exception,
            int codeBitsExpected,
            [CanBeNull] string startsWithExpected = null)
        {
            Assert.IsNotNull(exception);
            AssertUtilities.WriteLine($"exception: {exception.Message}");
            var serviceException = exception as UAServiceException;
            Assert.IsNotNull(serviceException);
            UAStatusCode statusCode = serviceException.ServiceResult.StatusCode;
            Assert.IsTrue(statusCode.IsBad);
            Assert.AreEqual(statusCode.CodeBits, codeBitsExpected);
            if (startsWithExpected != null)
                StringAssert.Contains(exception.Message, startsWithExpected);
        }

        [NotNull]
        protected IEnumerable<EasyUADataChangeNotificationEventArgs> CollectMonitoredItemChangedEventArgs(
            [NotNull] EasyUAClient client,
            [NotNull] EasyUAMonitoredItemArguments arguments,
            [NotNull] Action<int[]> actionWhileSubscribed)
        {
            IEnumerable<EasyUADataChangeNotificationEventArgs>[] resultArray = CollectMultipleDataChangeNotificationEventArgs(
                client, new[] { arguments }, actionWhileSubscribed);
            IEnumerable<EasyUADataChangeNotificationEventArgs> result = resultArray[0];
            Debug.Assert(result != null);
            return result;
        }

        [NotNull]
        protected IEnumerable<EasyUADataChangeNotificationEventArgs> CollectMonitoredItemChangedEventArgs(
            [NotNull] EasyUAClient client,
            [NotNull] EasyUAMonitoredItemArguments arguments,
            int millisecondsToCollect)
        {
            return CollectMonitoredItemChangedEventArgs(client, arguments, _ => AssertUtilities.WaitingPhase(millisecondsToCollect));
        }

        [NotNull]
        protected IEnumerable<EasyUADataChangeNotificationEventArgs>[] CollectMultipleDataChangeNotificationEventArgs(
            [NotNull] EasyUAClient client,
            [NotNull] EasyUAMonitoredItemArguments[] argumentsArray,
            [NotNull] Action<int[]> actionWhileSubscribed)
        {
            int length = argumentsArray.Length;
            var resultArray = new List<EasyUADataChangeNotificationEventArgs>[length];
            var easyArgumentsArray = new EasyUAMonitoredItemArguments[length];
            for (int i = 0; i < length; i++)
            {
                Debug.Assert(argumentsArray[i] != null);
                easyArgumentsArray[i] = new EasyUAMonitoredItemArguments(
                    (_, e) =>
                    {
                        Debug.Assert(e != null);
                        Debug.Assert(e.Arguments.State != null);
                        var j = (int)e.Arguments.State;
                        Debug.Assert(resultArray[j] != null);
                        resultArray[j].Add(e);
                    },
                    argumentsArray[i])
                {
                    State = i
                };
                resultArray[i] = new List<EasyUADataChangeNotificationEventArgs>();
            }


            int[] handles = client.SubscribeMultipleMonitoredItems(easyArgumentsArray);
            actionWhileSubscribed(handles);
            client.UnsubscribeMultipleMonitoredItems(handles);

            // ReSharper disable CoVariantArrayConversion
            return resultArray;
            // ReSharper restore CoVariantArrayConversion
        }

        [NotNull]
        protected IEnumerable<EasyUADataChangeNotificationEventArgs>[] CollectMultipleDataChangeNotificationEventArgs(
            [NotNull] EasyUAClient client,
            [NotNull] EasyUAMonitoredItemArguments[] argumentsArray,
            int millisecondsToCollect)
        {
            return CollectMultipleDataChangeNotificationEventArgs(client, argumentsArray,
                                                                _ => AssertUtilities.WaitingPhase(millisecondsToCollect));
        }

        [NotNull]
        protected EasyUAClient MakeClient()
        {
            AssertUtilities.WriteSeparatorLine();
            var client = new EasyUAClient
            {
                Isolated = true, 
                IsolatedParameters = EasyUAAdaptableParameters.OpcCompliance
            };
            //client.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.TrustEndpointUrlString(EndpointDescriptor.UrlString);
            client.IsolatedParameters.SessionParameters.HoldPeriod = 0;

            return client;
        }

        [NotNull]
        protected UAModelClient MakeModelClient()
        {
            var modelClient = UAModelClient.CreateWithServerModel();
            //modelClient.Client.IsolatedParameters.SessionParameters.CertificateAcceptancePolicy.TrustEndpointUrlString(EndpointDescriptor.UrlString);
            return modelClient;
        }

        [NotNull]
        protected EasyUAMonitoredItemArguments MakeMonitoredItemArguments([NotNull] Type type, bool dynamic)
        {
            var nodeDescriptor = UATestUtilities.MakeNodeDescriptor(type, dynamic);
            return new EasyUAMonitoredItemArguments(null, EndpointDescriptor, nodeDescriptor, 100);
        }

        [NotNull]
        protected EasyUAMonitoredItemArguments[] MakeMultipleMonitoredItemArguments([NotNull] IEnumerable<Type> types, bool dynamic)
        {
            return types.Select(type =>
                // ReSharper disable AssignNullToNotNullAttribute
                MakeMonitoredItemArguments(type, dynamic)).ToArray();
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [NotNull]
        protected EasyUAMonitoredItemArguments[] MakeVolumeMonitoredItemArguments(int count, bool dynamic)
        {
            return
                UATestUtilities.MakeVolumeNodeDescriptors(count, dynamic)
                    .Select(
                        nodeDescriptor =>
                            // ReSharper disable once AssignNullToNotNullAttribute
                            new EasyUAMonitoredItemArguments((EasyUADataChangeNotificationEventHandler)null, null, EndpointDescriptor, nodeDescriptor))
                    .ToArray();
        }

#endregion
    }
}
