﻿// $Header: $
// Copyright (c) CODE Consulting and Development, s.r.o., Plzen. All rights reserved.

using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.Win32;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.UA.Compliance
{
    class ScriptRunner : IDisposable
    {
        #region Private Constants
        
        private const int DelayAfterClose = 500;
        private const int DelayAfterStart = 1000;
        private const int DelayAfterKill = 500;
        private const int DelayBeforeClose = 0;

        #endregion



        #region Public Constructors

        public ScriptRunner([NotNull] string projectFile, [NotNull] string scriptFile, [NotNull] string resultFile)
        {
            ManualStart = false;
            //ManualClose = true;

            ProjectFile = projectFile;
            ScriptFile = scriptFile;
            ResultFile = resultFile;
        }

        #endregion



        #region Public Properties

        public bool ManualClose { get; set; }
        public bool ManualStart { get; set; }

        [NotNull]
        public string ProjectFile { get; }

        [NotNull]
        public string ResultFile { get; }

        [NotNull]
        public string ScriptFile { get; }

        #endregion



        #region Public Methods

        public void Run()
        {
            AssertUtilities.WriteLine($"Open CTT project: \"{ProjectFile}\" and run script: \"{ScriptFile}\"");

            if (ManualStart)
                return;

            RegistryKey localMachine = Registry.LocalMachine;
            Trace.Assert(localMachine != null);
            const string keyName = @"SOFTWARE\Microsoft\Windows\CurrentVersion\App Paths\uacompliancetest.exe";
            string appPath;
            using (RegistryKey key = localMachine.OpenSubKey(keyName))
            {
                if (key == null)
                    throw new ApplicationException("CTT application path key not found in the registry.");
                appPath = (string) key.GetValue(null);
                if (appPath == null)
                    throw new ApplicationException("CTT application path value not found in the registry.");
            }

            AssertUtilities.WriteLine($"CTT application path: {appPath}");
            FileVersionInfo fileVersionInfo = FileVersionInfo.GetVersionInfo(appPath);
            AssertUtilities.WriteLine($"CTT file version: {fileVersionInfo.FileVersion}");

            string fileName = appPath;
            string arguments = String.Format(CultureInfo.InvariantCulture,
                                                "--settings \"{0}\" " +
                                                "--file \"{1}\" " +
                                                "--result \"{2}\" " +
                                                "-k",   // --keepclienttrace, keep TRACE.DAT
                                                // TODO: -h, --hidden
                                                ProjectFile, ScriptFile, ResultFile);
            AssertUtilities.WriteLine($"CTT file name: {fileName}");
            AssertUtilities.WriteLine($"CTT arguments: {arguments}");
            var process = new Process
            {
                StartInfo = {FileName = fileName, Arguments = arguments}
            };
            if (!process.Start())
                throw new ApplicationException("The new CTT process has not been started.");
            CttProcess = process;

            Thread.Sleep(DelayAfterStart);   // give it some time to perform the script
        }

        public void Dispose()
        {
            AssertUtilities.WriteLine($"Finalize CTT script (\"{ScriptFile}\")");

            if (ManualClose)
                return;

            Process process = CttProcess;
            if (ManualStart)
            {
                Process[] processes = Process.GetProcessesByName("uacompliancetest");
                int length = processes.Length;
                if (length == 0)
                {
                    AssertUtilities.WriteLine("Could not find a CTT process to close!");
                    return;
                }
                if (length >= 2)
                    AssertUtilities.WriteLine("Found multiple CTT processes, closing first one of them!");
                process = processes[0];
            }

            Thread.Sleep(DelayBeforeClose);
            Trace.Assert(process != null);
            try
            {
                if (!process.CloseMainWindow())
                    AssertUtilities.WriteLine("Could not send a close message to the CTT main window!");
            }
            catch (InvalidOperationException)
            {
                AssertUtilities.WriteLine("The CTT process has already exited!");
            }
            if (!process.WaitForExit(DelayAfterClose))
            {
                AssertUtilities.WriteLine("The CTT process has not closed in time!");
                try
                {
                    process.Kill();
                }
                catch (Win32Exception win32Exception)
                {
                    AssertUtilities.WriteLine(win32Exception);
                }
                catch (InvalidOperationException invalidOperationException)
                {
                    AssertUtilities.WriteLine(invalidOperationException);
                }
                if (!process.WaitForExit(DelayAfterKill))
                    AssertUtilities.WriteLine("The CTT process has not exited in time!");
            }

            RaiseClosed();
        }

        #endregion



        #region Public Events

        [field: NonSerialized]
        public event EventHandler Closed;

        #endregion



        #region Private Properties

        [CanBeNull]
        private Process CttProcess { get; set; }
        #endregion



        #region Private Methods

        private void RaiseClosed()
        {
            EventHandler handler = Closed;
            handler?.Invoke(this, new EventArgs());
        }

        #endregion
    }
}
