﻿
using System;
using System.Diagnostics;
using System.Reactive.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Reactive;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.Reactive
{
    [TestClass]
    public class UAReactiveTest
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void DataChangeNotificationObservableTest()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            AssertUtilities.Phase("Creating observable");
            UADataChangeNotificationObservable<int> observable = UADataChangeNotificationObservable.Create<int>(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value,
                100);

            TestSection.Act();

            AssertUtilities.Phase("Subscribing");
            using (/*var subscription = */observable.Subscribe(e =>
                // ReSharper disable PossibleNullReferenceException
                AssertUtilities.WriteLine(e.TypedAttributeData.ToString())))
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.WaitingPhase(5 * 1000); // spin down

            TestSection.Assert();

            AssertUtilities.Done();
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void DataChangeNotificationObservableTest_InvalidCast()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            TestSection.Act();

            AssertUtilities.Phase("Creating observable");
            UADataChangeNotificationObservable<int> observable = UADataChangeNotificationObservable.Create<int>(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_Scalar_FloatValue,
                100);
            AssertUtilities.Phase("Subscribing");
            using (/*var subscription = */observable.Subscribe(AssertUtilities.WriteLine))
                AssertUtilities.WaitingPhase(10 * 1000);

            TestSection.Assert();
            AssertUtilities.Done();
        }

        [TestCategory("Live")]
        [TestCategory("OpcUA")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void PipelineTest()
        {
            TestSection.Arrange();
            UAEndpointDescriptor endpointDescriptor = UATestEnvironment.OpcFoundation_UASampleServer.EndpointDescriptors.OpcTcp;
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            UADataChangeNotificationObservable<int> sourceObservable = UADataChangeNotificationObservable.Create<int>(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Dynamic_UserScalar_Int32Value,
                100);
            IObservable<int> processedObservable = sourceObservable.Select(e =>
                // ReSharper disable PossibleNullReferenceException
                e.TypedAttributeData.TypedValue % 1000);
                // ReSharper restore PossibleNullReferenceException
            Debug.Assert(processedObservable != null);

            UAWriteValueObserver<int> observer = UAWriteValueObserver.Create<int>(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value);
            UADataChangeNotificationObservable<int> destinationObservable = UADataChangeNotificationObservable.Create<int>(
                endpointDescriptor,
                OpcFoundation_UASampleServer.NodeIdExpandedTexts.Data_Static_UserScalar_Int32Value,
                100);

            TestSection.Act();

            using (/*var sourceSubscription = */processedObservable.Subscribe(observer))
            using (/*var destinationSubscription = */destinationObservable.Subscribe(e =>
                // ReSharper disable PossibleNullReferenceException
                AssertUtilities.WriteLine(e.TypedAttributeData.ToString())))
                // ReSharper restore PossibleNullReferenceException
                AssertUtilities.WaitingPhase(10 * 1000);

            // Spinning down is absolutely necessary here, as the running pipeline writes into nodes that we are using in
            // other tests, and we must prevent it from possibly spoiling them.
            AssertUtilities.WaitingPhase(5*1000);   // spin down

            TestSection.Assert();
        }

        #endregion



        #region Private Properties

        [NotNull]
        private UATestEnvironment TestEnvironment => UATestEnvironment.Current;

        #endregion
    }
}
