﻿
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Diagnostics;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Static Private Methods

        [NotNull]
        static private StringObjectDictionary CollectBasicInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary
            {
                ["BuildInfo"] = CollectBuildInfoInfo(client, endpointDescriptor),
                ["Capabilities"] = CollectCapabilitiesInfo(client, endpointDescriptor),
                ["DataTypeModel"] = CollectDataTypeModelInfo(client, endpointDescriptor),
                //["EndpointDescriptor.UrlString"] = endpointDescriptor.UrlString,  // rather not
                ["HistoryServer"] = CollectHistoryServerInfo(client, endpointDescriptor),
                ["InformationModel"] = CollectInformationModelInfo(client, endpointDescriptor),
                ["OperationLimits"] = CollectOperationLimitsInfo(client, endpointDescriptor),
                ["PubSub"] = CollectPubSubInfo(client, endpointDescriptor),
                ["Redundancy"] = CollectRedundancyInfo(client, endpointDescriptor),
                ["Server"] = CollectServerInfo(client, endpointDescriptor),
            };
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectBuildInfoInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                BuildInfo buildInfo = client.ReadValueNonNull<BuildInfo>(endpointDescriptor, UAVariableIds.Server_ServerStatus_BuildInfo);
                result["BuildDate"] = buildInfo.BuildDate.ToString(CultureInfo.InvariantCulture);
                result["BuildNumber"] = buildInfo.BuildNumber;  // a String
                result["ManufacturerName"] = buildInfo.ManufacturerName;
                result["ProductName"] = buildInfo.ProductName;
                result["ProductUri"] = buildInfo.ProductUri;
                result["SoftwareVersion"] = buildInfo.SoftwareVersion;  // a String
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectCapabilitiesInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                result["LocaleIds"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_LocaleIdArray);
                result["MaxArrayLength"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MaxArrayLength);
                result["MaxBrowseContinuationPoints"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MaxBrowseContinuationPoints);
                result["MaxHistoryContinuationPoints"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MaxHistoryContinuationPoints);
                result["MaxQueryContinuationPoints"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MaxQueryContinuationPoints);
                result["MaxStringLength"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MaxStringLength);
                result["MinSupportedSampleRate"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_MinSupportedSampleRate);
                result["ServerProfiles"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerCapabilities_ServerProfileArray);
                // IMPROVE: SoftwareCertificates
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectDataTypeModelInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();

            ValueResult<UANodeDescriptor[]> dataTypeArrayResult = client.RetrieveAllDataTypes(endpointDescriptor);
            if (dataTypeArrayResult.Succeeded)
            {
                UANodeDescriptor[] dataTypeArray = dataTypeArrayResult.Value;
                Debug.Assert(dataTypeArray != null);
                result["DataTypes.Count"] = dataTypeArray.Length;
                IEnumerable<UANodeDescriptor> nonStandardDataTypes = dataTypeArray.Where(descriptor =>
                    // ReSharper disable once PossibleNullReferenceException
                    descriptor.NodeId.NamespaceUriString != UANamespaceQualified.StandardNamespaceUriString);
                int nonStandardDataTypeCount = nonStandardDataTypes.Count();
                result["NonStandardDataTypes.Count"] = nonStandardDataTypeCount;
                //foreach (UANodeDescriptor nonStandardDataType in nonStandardDataTypes)
                //    AssertUtilities.WriteLine($"{nonStandardDataType.BrowsePath} -> {nonStandardDataType.NodeId}");

                // ReSharper disable once PossibleNullReferenceException
                var dataTypeNamespaceSet = new HashSet<string>(dataTypeArray.Select(descriptor => descriptor.NodeId.NamespaceUriString));
                result["DataTypeNamespaces"] = dataTypeNamespaceSet.ToList();

                ValueResult[] dataTypeDefinitionResultArray = client.ReadMultipleValues(dataTypeArray.Select(descriptor =>
                    // ReSharper disable once AssignNullToNotNullAttribute
                    new UAReadArguments(endpointDescriptor, descriptor, UAAttributeId.DataTypeDefinition)).ToArray());

                // ReSharper disable once PossibleNullReferenceException
                int succeededCount = dataTypeDefinitionResultArray.Count(valueResult => valueResult.Succeeded);
                result["DataTypeDefinitions.Count"] = succeededCount;
            }

            ValueResult dataTypeDefinitionResult = client.ReadMultipleValues(new[]
            {
                new UAReadArguments(
                    endpointDescriptor,
                    UADataTypeIds.ServerStatusDataType,
                    UAAttributeId.DataTypeDefinition)
            })[0];
            Check.AssertNotNull(dataTypeDefinitionResult);
            bool? dataTypeDefinitionAttribute = null;
            if (dataTypeDefinitionResult.Succeeded)
                dataTypeDefinitionAttribute = true;
            else if ((dataTypeDefinitionResult.Exception is UAStatusCodeException statusCodeException) &&
                     (statusCodeException.StatusCode.CodeBits == UACodeBits.BadAttributeIdInvalid))
                dataTypeDefinitionAttribute = false;
            result["DataTypeDefinitionAttribute"] = dataTypeDefinitionAttribute.HasValue ? (object)dataTypeDefinitionAttribute.Value : null;

            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectHistoryServerInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                result["AccessHistoryDataCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_AccessHistoryDataCapability);
                result["AccessHistoryEventsCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_AccessHistoryEventsCapability);
                result["DeleteAtTimeCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_DeleteAtTimeCapability);
                result["DeleteEventCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_DeleteEventCapability);
                result["DeleteRawCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_DeleteRawCapability);
                result["InsertAnnotationCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_InsertAnnotationCapability);
                result["InsertDataCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_InsertDataCapability);
                result["InsertEventCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_InsertEventCapability);
                result["MaxReturnDataValues"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_MaxReturnDataValues);
                result["MaxReturnEventValues"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_MaxReturnEventValues);
                result["ReplaceDataCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_ReplaceDataCapability);
                result["ReplaceEventCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_ReplaceEventCapability);
                result["UpdateDataCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_UpdateDataCapability);
                result["UpdateEventCapability"] = client.ReadValue(endpointDescriptor, UAVariableIds.HistoryServerCapabilities_UpdateEventCapability);
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectInformationModelInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                string[] namespaceArray = client.ReadValueNonNull<string[]>(endpointDescriptor, UAVariableIds.Server_NamespaceArray);
                result["OpcUa.NamespaceListed"] = namespaceArray.Contains(UAUtilities.OpcUANamespaceUriString);
#if (NETFRAMEWORK)    // TODO:
                result["OpcUaAdi.NamespaceListed"] = namespaceArray.Contains(Opc.Ua.Adi.Namespaces.OpcUaAdi);
#endif
                // ReSharper disable once StringLiteralTypo
                result["OpcUaBACnet.NamespaceListed"] = namespaceArray.Contains("http://opcfoundation.org/UA/BACnet/ ");    //?
#if (NETFRAMEWORK)    // TODO:
                result["OpcUaDi.NamespaceListed"] = namespaceArray.Contains(Opc.Ua.Di.Namespaces.OpcUaDi);
                result["OpcUaFdi5.NamespaceListed"] = namespaceArray.Contains(Opc.Ua.Fdi5.Namespaces.OpcUaFdi5);
#endif
                result["OpcUaGds.NamespaceListed"] = namespaceArray.Contains(Opc.Ua.Gds.Namespaces.OpcUaGds);
                result["OpcUaIsa95.NamespaceListed"] = namespaceArray.Contains("http://www.OPCFoundation.org/UA/2013/01/ISA95");
                result["OpcUaMTConnect.NamespaceListed"] = namespaceArray.Contains("http://opcfoundation.org/UA/MTConnect/");
#if (NETFRAMEWORK)    // TODO:
                result["OpcUaPlc.NamespaceListed"] = namespaceArray.Contains(Opc.Ua.Plc.Namespaces.OpcUaPlc);
#endif
                // ReSharper disable once StringLiteralTypo
                result["OpcUaPowerlink.NamespaceListed"] = namespaceArray.Contains("http://opcfoundation.org/UA/POWERLINK/");
                result["OpcUaTmc.NamespaceListed"] = namespaceArray.Contains("http://opcfoundation.org/UA/TMC/");
                // ReSharper disable once StringLiteralTypo
#if (NETFRAMEWORK)    // TODO:
                result["Sercos.NamespaceListed"] = namespaceArray.Contains(Sercos.Namespaces.Sercos);
#endif
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectOperationLimitsInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                result["MaxMonitoredItemsPerCall"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxMonitoredItemsPerCall);
                result["MaxNodesPerBrowse"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerBrowse);
                result["MaxNodesPerHistoryReadData"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerHistoryReadData);
                result["MaxNodesPerHistoryReadEvents"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerHistoryReadEvents);
                result["MaxNodesPerHistoryUpdateData"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerHistoryUpdateData);
                result["MaxNodesPerHistoryUpdateEvents"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerHistoryUpdateEvents);
                result["MaxNodesPerMethodCall"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerMethodCall);
                result["MaxNodesPerNodeManagement"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerNodeManagement);
                result["MaxNodesPerRead"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerRead);
                result["MaxNodesPerRegisterNodes"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerRegisterNodes);
                result["MaxNodesPerTranslateBrowsePathsToNodeIds"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerTranslateBrowsePathsToNodeIds);
                result["MaxNodesPerWrite"] = client.ReadValue(endpointDescriptor, UAVariableIds.OperationLimitsType_MaxNodesPerWrite);
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectPubSubInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                client.ReadValue(endpointDescriptor, UAObjectIds.PublishSubscribe, UAAttributeId.NodeClass);
                result["PublishSubscribe"] = null;
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectRedundancyInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                result["RedundancySupport"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerRedundancy_RedundancySupport);
                result["ServerUris"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerRedundancy_ServerUriArray);
            }
            catch (UAException)
            {
            }
            return result;
        }

        [NotNull]
        static private StringObjectDictionary CollectServerInfo(
            [NotNull] IEasyUAClient client,
            [NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            var result = new StringObjectDictionary();
            try
            {
                result["Auditing"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_Auditing);
                result["Namespaces"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_NamespaceArray);
                result["Servers"] = client.ReadValue(endpointDescriptor, UAVariableIds.Server_ServerArray);
            }
            catch (UAException)
            {
            }
            return result;
        }

#endregion



#region Test Methods

        [TestMethod]
        public void Group101_Features_Header____________________()
        {
        }

        [TestMethod]
        public void Group101_Features_Test01_Basic()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            AssertUtilities.WriteLine($"Endpoint descriptor: {endpointDescriptor}");

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                StringObjectDictionary basicInfo = CollectBasicInfo(client, endpointDescriptor);
                IDictionary<string, object> flattenedBasicInfo = basicInfo.Flatten((a, b) => a + "." + b);

                AssertUtilities.WriteLine();
                foreach (KeyValuePair<string, object> pair in flattenedBasicInfo)
                {
                    object value = pair.Value;
                    string sValue = value is IList list ? string.Join(", ", list.Cast<object>().ToArray()) : value?.ToString();
                    AssertUtilities.WriteLine($"{pair.Key}: {sValue}");
                }

                TestSection.Assert();
            }
        }

#endregion
    }
}
